==> libso-2007.2.tgz: libso/src/mapping.c <==

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/file.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include "so_internal_arch.h"
#include "so_qlock.h"
#include "litmus_so.h"

lso_global_t __lso_malloc(struct litmus_so* lso, unsigned int size);
lso_global_t __allocate_dictionary(struct litmus_so* lso);

static int map_file(const char* filename, void **addr, size_t *size, int *desc)
{
	struct stat info;
	int error = 0;
	int fd;

	error = stat(filename, &info);
	if (!error) {
		*size = info.st_size;
		if (info.st_size > 0) {
			fd = open(filename, O_RDWR);
			if (fd >= 0) {
				*addr = mmap(NULL, *size, 
					     PROT_READ | PROT_WRITE, MAP_SHARED,
					     fd, 0);
				if (*addr == MAP_FAILED)
					error = -1;
				*desc = fd;
			} else
				error = fd;
		} else
			*addr = NULL;
	}
	return error;
}


static int allocate_pid(char* pidmap, int length)
{
	int i;
	char old;
	for (i = 0; i < length; i++) {
		old = test_and_set_char(1, pidmap + i);
		if (!old)
			return i;
	}
	return -1;
}


lso_error_t lso_open(const char* file, lso_t* handle)
{
	int err;
	struct litmus_so* lso = NULL;
	lso_error_t error = LSO_IO_ERROR;
	size_t mapped_size;
	char* pidmap;

	lso = malloc(sizeof(struct litmus_so));
	if (!lso)
		goto cleanup;
	
	lso->base = NULL;
	err = map_file(file, (void**) &lso->base, &mapped_size, &lso->fd);
	if (err)
		goto cleanup;

	if (lso->base->magic1 != LSO_MAGIC_1 ||
	    lso->base->magic2 != LSO_MAGIC_2) {
		error = LSO_NO_OBJECT;
		goto cleanup;
	}

	if (lso->base->size != mapped_size) {
		error = LSO_INCONSISTENT;
		goto cleanup;
	}

	/* ok, it seems we have mapped a shared object 
	 * let's try to obtain a PID, eg. check whether not too many processes
	 * have mapped the object already
	 */

	pidmap = (char*) lso_local_ptr((lso_t) lso, lso->base->pid_map);
	lso->pid = allocate_pid(pidmap, lso->base->max_pid);

	if (lso->pid == -1)
		/* too many already! */
		error = LSO_NO_PID;
	else {
		/* we got one! */
		*handle  = (lso_t) lso;
		error    = LSO_OK;
	}

 cleanup:
	if (error != LSO_OK && lso) {
		if (lso->base)
			munmap(lso->base, mapped_size);
		free(lso);
	}
	return error;
}


lso_error_t lso_close(lso_t* handle)
{
	struct litmus_so* lso = (struct litmus_so*) *handle;
	char* pidmap;
	size_t size;

	*handle = NULL;

	/* release PID */
	pidmap = (char*) lso_local_ptr((lso_t) lso, lso->base->pid_map);
	*(pidmap + lso->pid) = 0;

	/* remove object from address space*/
	size = lso->base->size;
	munmap(lso->base, size);

	/* free management info */
	free(lso);

	return LSO_OK;
}


/*
 * TODO: - handle the case for different mount points where link will fail
 */
lso_error_t lso_create(const char* file, unsigned int number_of_procs, 
		       unsigned int heap_size)
{
	int fd, err;
	lso_error_t error = LSO_IO_ERROR;
	size_t size;
	struct litmus_so 	lso;
	struct lso_heap_cell*	cell;
	char template[10];

	if (heap_size < 1024 * number_of_procs)
		heap_size = 1024 * number_of_procs;

	lso.base = NULL;
	strncpy(template, "lsoXXXXXX", 10);

	/* create a private file first so nobody can observe the
	 * object while it is being created
	 */
	fd = mkstemp(template);
	if (fd < 0) {
		if (errno == EEXIST)
			return LSO_EXISTS;
		else
			return LSO_IO_ERROR;
	}

	size = sizeof(struct lso_header) + heap_size;
	err = ftruncate(fd, size);
	if (err < 0) 
		goto io_error;

	lso.pid = 0;		
	lso.base = mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
	if (lso.base == MAP_FAILED) {
		lso.base = NULL;
		goto io_error;
	}

	lso.base->max_pid  = number_of_procs;
	lso.base->size     = size;

	/* bootstrap memory management */
	lso.base->heap.start = sizeof(struct lso_header);
	lso.base->heap.end   = size;
	lso.base->heap.free_list_head = lso.base->heap.start;
	
	cell = (struct lso_heap_cell*) (((char*) lso.base) + lso.base->heap.start);
	cell->size = size - sizeof(struct lso_header) - sizeof(struct lso_heap_cell);
	cell->next = lso.base->heap.start;
	cell->prev = lso.base->heap.start;

	/* allocate PID map                */
	lso.base->pid_map = __lso_malloc(&lso, sizeof(char) * number_of_procs);
	if (!lso.base->pid_map) {
		error = LSO_NO_MEM;
		goto io_error;
	}

	/* allocate and setup dictionary   */
	__allocate_dictionary(&lso);	

	/* mark as LSO */
	lso.base->magic1 = LSO_MAGIC_1;
	lso.base->magic2 = LSO_MAGIC_2;

	/* ok, setup is complete. Now let's try to make it known as the real
	 * object
	 */
	err = link(template, file);
	if (err < 0) {
		if (errno == EEXIST)
			error = LSO_EXISTS;
	} else
		/* the file is setup and was moved to the correct location */
		/* shared object creation is done */
		error = LSO_OK;

 io_error:
	if (lso.base)
		munmap(lso.base, size);
	close(fd);
	unlink(template);
	return error;
}


void __lso_lock(struct litmus_so* lso)
{
	flock(lso, LOCK_EX);
}

void __lso_unlock(struct litmus_so* lso)
{
	flock(lso, LOCK_UN);
}

==> libso-2007.3.tgz: libso/src/mapping.c <==

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/file.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include "so_internal_arch.h"
#include "so_qlock.h"
#include "litmus_so.h"

lso_global_t __lso_malloc(struct litmus_so* lso, unsigned int size);
lso_global_t __allocate_dictionary(struct litmus_so* lso);

static int map_file(const char* filename, void **addr, size_t *size, int *desc)
{
	struct stat info;
	int error = 0;
	int fd;

	error = stat(filename, &info);
	if (!error) {
		*size = info.st_size;
		if (info.st_size > 0) {
			fd = open(filename, O_RDWR);
			if (fd >= 0) {
				*addr = mmap(NULL, *size, 
					     PROT_READ | PROT_WRITE, MAP_SHARED,
					     fd, 0);
				if (*addr == MAP_FAILED)
					error = -1;
				*desc = fd;
			} else
				error = fd;
		} else
			*addr = NULL;
	}
	return error;
}


static int allocate_pid(char* pidmap, int length)
{
	int i;
	char old;
	for (i = 0; i < length; i++) {
		old = test_and_set_char(1, pidmap + i);
		if (!old)
			return i;
	}
	return -1;
}


lso_error_t lso_open(const char* file, lso_t* handle)
{
	int err;
	struct litmus_so* lso = NULL;
	lso_error_t error = LSO_IO_ERROR;
	size_t mapped_size;
	char* pidmap;

	lso = malloc(sizeof(struct litmus_so));
	if (!lso)
		goto cleanup;
	
	lso->base = NULL;
	err = map_file(file, (void**) &lso->base, &mapped_size, &lso->fd);
	if (err)
		goto cleanup;

	if (lso->base->magic1 != LSO_MAGIC_1 ||
	    lso->base->magic2 != LSO_MAGIC_2) {
		error = LSO_NO_OBJECT;
		goto cleanup;
	}

	if (lso->base->size != mapped_size) {
		error = LSO_INCONSISTENT;
		goto cleanup;
	}

	/* ok, it seems we have mapped a shared object 
	 * let's try to obtain a PID, eg. check whether not too many processes
	 * have mapped the object already
	 */

	pidmap = (char*) lso_local_ptr((lso_t) lso, lso->base->pid_map);
	lso->pid = allocate_pid(pidmap, lso->base->max_pid);

	if (lso->pid == -1)
		/* too many already! */
		error = LSO_NO_PID;
	else {
		/* we got one! */
		*handle  = (lso_t) lso;
		error    = LSO_OK;
	}

 cleanup:
	if (error != LSO_OK && lso) {
		if (lso->base)
			munmap(lso->base, mapped_size);
		free(lso);
	}
	return error;
}


lso_error_t lso_close(lso_t* handle)
{
	struct litmus_so* lso = (struct litmus_so*) *handle;
	char* pidmap;
	size_t size;

	*handle = NULL;

	/* release PID */
	pidmap = (char*) lso_local_ptr((lso_t) lso, lso->base->pid_map);
	*(pidmap + lso->pid) = 0;

	/* remove object from address space*/
	size = lso->base->size;
	munmap(lso->base, size);

	/* free management info */
	free(lso);

	return LSO_OK;
}


/*
 * TODO: - handle the case for different mount points where link will fail
 */
lso_error_t lso_create(const char* file, unsigned int number_of_procs, 
		       unsigned int heap_size)
{
	int fd, err;
	lso_error_t error = LSO_IO_ERROR;
	size_t size;
	struct litmus_so 	lso;
	struct lso_heap_cell*	cell;
	char template[10];

	if (heap_size < 1024 * number_of_procs)
		heap_size = 1024 * number_of_procs;

	lso.base = NULL;
	strncpy(template, "lsoXXXXXX", 10);

	/* create a private file first so nobody can observe the
	 * object while it is being created
	 */
	fd = mkstemp(template);
	if (fd < 0) {
		if (errno == EEXIST)
			return LSO_EXISTS;
		else
			return LSO_IO_ERROR;
	}

	size = sizeof(struct lso_header) + heap_size;
	err = ftruncate(fd, size);
	if (err < 0) 
		goto io_error;

	lso.pid = 0;		
	lso.base = mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
	if (lso.base == MAP_FAILED) {
		lso.base = NULL;
		goto io_error;
	}

	lso.base->max_pid  = number_of_procs;
	lso.base->size     = size;

	/* bootstrap memory management */
	lso.base->heap.start = sizeof(struct lso_header);
	lso.base->heap.end   = size;
	lso.base->heap.free_list_head = lso.base->heap.start;
	
	cell = (struct lso_heap_cell*) (((char*) lso.base) + 
					lso.base->heap.start);
	cell->size = size - sizeof(struct lso_header) - 
		sizeof(struct lso_heap_cell);
	cell->next = lso.base->heap.start;
	cell->prev = lso.base->heap.start;

	/* allocate PID map                */
	lso.base->pid_map = __lso_malloc(&lso, sizeof(char) * number_of_procs);
	if (!lso.base->pid_map) {
		error = LSO_NO_MEM;
		goto io_error;
	}

	/* allocate and setup dictionary   */
	__allocate_dictionary(&lso);	

	/* mark as LSO */
	lso.base->magic1 = LSO_MAGIC_1;
	lso.base->magic2 = LSO_MAGIC_2;

	lso.base->seq_no = 1;

	/* ok, setup is complete. Now let's try to make it known as the real
	 * object
	 */
	err = link(template, file);
	if (err < 0) {
		if (errno == EEXIST)
			error = LSO_EXISTS;
	} else
		/* the file is setup and was moved to the correct location */
		/* shared object creation is done */
		error = LSO_OK;

 io_error:
	if (lso.base)
		munmap(lso.base, size);
	close(fd);
	unlink(template);
	return error;
}


void __lso_lock(struct litmus_so* lso)
{
	flock(lso, LOCK_EX);
}

void __lso_unlock(struct litmus_so* lso)
{
	flock(lso, LOCK_UN);
}
